from ..GeneSets.geneSetObjects import getGSNA
from ..GeneSets.geneSetScores import GeneSetScore

from ..GeneExpressions.geneExpScores import expScore

import numpy as np

from anndata import AnnData


def runGSNA(adata, jsonFile, normalized=False):
    # TODO : Print normalization process info.

    scoresArray = []
    geneSetNamesArray = []
    # Create arrays to store scores and names of gene sets

    geneSetList = getGSNA(jsonFile)
    # Create GeneSet objects from the JSON file.

    for geneset in geneSetList:
        # Calculate scores for each gene set.

        geneSetNamesArray.append(geneset.getID)

        _geneNames = geneset.getGeneNames

        newGeneSetScore = GeneSetScore(geneset.matrix, _geneNames)

        newExpScore = expScore(adata, newGeneSetScore)

        scoresArray.append(newExpScore)

    scoresArray = np.array(scoresArray).T

    adataScores = AnnData(X=scoresArray, var=geneSetNamesArray,
                          obs=adata.obs)

    if normalized:
        # If the normalized option is on, normalize the score data with
        # Quantile normalization and Z-Score normalization.

        from sklearn.preprocessing import quantile_transform, StandardScaler

        adataScores.X = quantile_transform(adataScores.X, axis=1,
                                           output_distribution="normal")

        adataScores.X = StandardScaler().fit_transform(adataScores.X)

    return adataScores



from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
import pandas as pd
from concurrent.futures import ThreadPoolExecutor
import requests
import json
import gc


def create_session():
    session = requests.Session()
    retries = Retry(total=5, backoff_factor=0.2, status_forcelist=[500, 502, 503, 504])
    adapter = HTTPAdapter(max_retries=retries)
    session.mount("https://", adapter)
    return session


def fetch_filtered_string_network(genes,
                                  session,
                                  species="9606",
                                  required_score=100,
                                  limit=1000):

    api_url = "https://string-db.org/api/json/network"
    data = {
        "identifiers": "\n".join(genes),
        "species": species,
        "required_score": required_score,
        "limit": limit,
        "network_type": "functional"
    }
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    
    try:
        response = session.post(api_url, data=data, headers=headers)
        response.raise_for_status()
        network_data = response.json()
        if not network_data:
            print(f"Warning! The Server Returned Null. Genes: {genes}")
            return None
        df = pd.DataFrame(network_data)
        filtered_df = df[(df['preferredName_A'].isin(genes)) & (df['preferredName_B'].isin(genes))]
        return filtered_df.to_dict(orient="records")
    except requests.exceptions.RequestException as e:
        print(f"Data could not be pulled error : {e}")
        return None


def process_gene_set(index, gene_set_name, genes, session):
    print(f"Iteration {index + 1}: Gene Set {gene_set_name}...")
    result = fetch_filtered_string_network(genes, session, limit=len(genes))
    gc.collect()
    return gene_set_name, result


def fetch_all_networks_parallel(gene_sets, max_workers=25):
    results = {}

    with create_session() as session:
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = {
                executor.submit(process_gene_set, index, gene_set_name, 
                                genes['geneSymbols'], session): gene_set_name

                for index,
                (gene_set_name, genes) in enumerate(gene_sets.items())
            }

            for future in futures:
                gene_set_name, result = future.result()
                results[gene_set_name] = result

    return results

# TODO: FARKLI DOSYA BICIMLERI EKLE


def parse_gmt(file_path):
    gene_sets = {}
    with open(file_path, "r") as f:
        for line in f:
            parts = line.strip().split("\t")
            if len(parts) < 3:
                continue
            gene_set_name = parts[0]
            gene_symbols = parts[2:]
            gene_sets[gene_set_name] = {'geneSymbols': gene_symbols}
    return gene_sets


def parse_tsv(file_path):
    gene_sets = {}
    with open(file_path, "r") as f:
        for line in f:
            parts = line.strip().split("\t")
            if len(parts) < 2:
                continue
            gene_set_name = parts[0]
            gene_symbols = parts[1:]
            gene_sets[gene_set_name] = {'geneSymbols': gene_symbols}
    return gene_sets


def makeJson(msigdbFile, fileType, jsonFileName="geneSets.json"):

    # if fileType == "json":

    # gene_sets = json.load(f)

    if fileType == "gmt":
        gene_sets = parse_gmt(msigdbFile)
    elif fileType == "tsv":
        gene_sets = parse_tsv(msigdbFile)
    else:
        raise ValueError(f"Unsupported file type: {fileType}")

    all_networks = fetch_all_networks_parallel(gene_sets, max_workers=40)

    relation_dict = {}

    for gene_set_id, interactions in all_networks.items():
        setBuffer = set()
        for interaction in interactions:
            node1 = interaction['preferredName_A']
            node2 = interaction['preferredName_B']
            combined_score = interaction['score']

            setBuffer.add(node1)
            setBuffer.add(node2)

            if gene_set_id not in relation_dict:
                relation_dict[gene_set_id] = {}

            if node1 not in relation_dict[gene_set_id]:
                relation_dict[gene_set_id][node1] = {}
            relation_dict[gene_set_id][node1][node2] = combined_score

            if node2 not in relation_dict[gene_set_id]:
                relation_dict[gene_set_id][node2] = {}
            relation_dict[gene_set_id][node2][node1] = combined_score

        for i in gene_sets[gene_set_id]['geneSymbols']:
            if i not in setBuffer:
                relation_dict[gene_set_id][i] = {}

    with open(jsonFileName, "w") as f:
        json.dump(relation_dict, f, indent=2)



from .GeneExpressions.geneExpScores import expScore

from .GeneSets.geneSetObjects import GeneSetMatrix, GeneSet, getGSNA
from .GeneSets.geneSetScores import GeneSetScore

from .Utils.jsonParser import pjson
from .Utils.MatrixItem import MatrixItem
from .Utils._annData import runGSNA
from .Utils.makeJsonFile import makeJson



import json


class pjson(dict):

    def __init__(self, jsonFile):

        self.jsonFile = jsonFile
        self.geneSayi = []
        self.geneNamesList = []
        self.iter_count = 0

        with open(jsonFile) as f:
            # Load the json file
            self.js = json.load(f)

        for i in self.js.values():

            for j in i.keys():
                self.geneNamesList.append(j)
                # Get the gene names

            self.geneSayi.append(len(i))
            # Get the gene counts

        self.update(self.js)

    def genesets(self):
        # Return genesets in the JSON file,
        # since they are represented in self.values().
        # Used for easier iteration.

        return self.values()

    @property
    def getAsDict(self):
        return self.js

    @property
    def getGeneNames(self):
        # Acces the gene names
        return self.geneNamesList

    @property
    def getGeneCounts(self):
        # Access the gene counts
        return self.geneSayi

    @property
    def getGeneSetCount(self):
        # Access the gene set count
        return len(self.geneSayi)

    @property
    def getFileInfo(self):
        # Access the file info as a string
        info = ""

        for i, j in enumerate(self.geneSayi):
            info += f"GeneSet : {i+1}, Gene Count : {j} \n"

        return info

    def getFileName(self):
        return self.jsonFile



import numpy as np

from ..Utils.MatrixItem import MatrixItem


class GeneSetMatrix:

    """ Creates and returns the gene set as an incidence matrix. 


        If we think of the gene sets and the relations between genes, they are 
            
        very similar to graphs. 

        The best way to represent graphs mathematically is by using an 
            
        incidence matrix. With incidence matrixes,
            
        our mathematical formula can be solved 
            
        for each geneset expeditiously.

        
        The main idea is to create an incidence matrix with specific positions 
        
        for each gene to make calculations
        
        efficiently. In order to do this, 
        
        we give each gene a specific column value.

    """
    
    def __new__(cls, rawGeneSet):

        unique_identifiers = {}
        # The unique identifiers dictionary to hold the data of the unique
        # position of an edge(weight, relation whetever you wanna call it)
        # in the matrix for faster calculations.

        n1 = 0

        position_in_column = 0
        GeneList = []

        for i in rawGeneSet:
            n1 += 1

            for j in rawGeneSet[i]:

                unique_hash = hash(i) * hash(j) * hash(rawGeneSet[i][j])
                # Creating a unique hash to determine the
                # unique position.

                if unique_hash not in unique_identifiers:
                    # If the edge is unique, give it a new position by giving
                    # it a new column value.

                    unique_identifiers[unique_hash] = position_in_column
                    position_in_column += 1

                GeneList.append(MatrixItem(float(rawGeneSet[i][j]), n1-1,
                                           unique_identifiers[unique_hash]-1))
                # Whether the edge is unique or not, create a MatrixItem
                # to hold the position data and add it to
                # the GeneList.

        n = len(unique_identifiers)

        matrix = np.zeros((n1, n))
        # Create the matrix. TODO: Test the sparse matrix approach.

        for item in GeneList:
            # Build the incidence matrix using
            # the position data in the GeneList.

            w, r, c = item.weight, item.row, item.column
            matrix[r, c] = w

        return matrix


class GeneSet:

    """ Holds the properties of a gene set. """

    def __init__(self, _id, _rawGeneSet):

        self.id = _id
        self.asJson = _rawGeneSet
        self.matrix = GeneSetMatrix(_rawGeneSet)

    @property
    def getID(self):

        return self.id

    @property
    def getAsJson(self):

        return self.asJson

    @property
    def getGeneNames(self):
        # Access gene names as a list

        namesListBuffer = [i for i in self.asJson]

        return namesListBuffer


def getGSNA(jsonFile):
    # Automatically create geneset objects and return them
    # in a list.

    _list = []

    for i in jsonFile:

        newGeneSet = GeneSet(f"{i}", jsonFile[i])

        _list.append(newGeneSet)

    return _list



import numpy as np


class GeneSetScore(dict):

    def __init__(self, matrix, geneNamesList):

        self.matrix = matrix

        self.geneNamesList = geneNamesList

        num_rows, num_cols = matrix.shape

        row_nz_counts = np.count_nonzero(matrix, axis=1)
        # Get the non-zero element count for each row

        scores = np.zeros(num_rows)
        # Create an empty matrix to hold the scores.

        for col in range(num_cols):
            nonzero_rows = np.nonzero(matrix[:, col])[0]
            # Get nonzero rows.

            if nonzero_rows.size == 2:
                # All columns in the GeneSetMatrix object are gonna have only 2
                # elements that are different from 0,
                # control it just in case.

                i, j = nonzero_rows
                # Get the two elements in a column to i,j

                scores[i] += matrix[i, col] * row_nz_counts[j]
                # Multiply the value in the row i
                # with the nonzero count in the row j and
                # add that to the score of row i.

                scores[j] += matrix[j, col] * row_nz_counts[i]
                # Vice versa.

        for i, gene in enumerate(geneNamesList):
            # Create the dictionary
            self[gene] = scores[i]



from dataclasses import dataclass


@dataclass
class MatrixItem:
    # MatrixItem class. Holds all of the data from the json entries, and the
    # location in the matrix.
    weight: float
    row: int
    column: int



import numpy as np


class GeneSetScore(dict):

    def __init__(self, matrix, geneNamesList, epsilon=10**-6):

        self.epsilon = epsilon

        self.matrix = matrix

        self.geneNamesList = geneNamesList

        for gene in self.geneNamesList:
            # Create the empty dictionary
            # Which all the scores within are initially zero.
            self[gene] = 0

        num_rows, num_cols = matrix.shape

        row_nz_counts = np.count_nonzero(matrix, axis=1)
        # Get the non-zero element count for each row

        disconnected_genes = np.where(row_nz_counts == 0)[0]
        for idx in disconnected_genes:

            self[geneNamesList[idx]] += self.epsilon

        # Search for rows that contain only zeros,
        # Those are the genes that are not related to
        # any other gene. If you find one, add epsilon to
        # the score of the respected gene, which is 0.
        # This way, genes that are not in relation to others
        # can have an effect to the score.

        for col in range(num_cols):
            i, j = np.nonzero(matrix[:, col])[0]
            # Get nonzero rows as i,j

            self[geneNamesList[i]] += matrix[i, col] * row_nz_counts[j]
            # Multiply the value in the row i
            # with the nonzero count in the row j and
            # add that to the score of row i.

            self[geneNamesList[j]] += matrix[j, col] * row_nz_counts[i]
            # Vice versa.

       



import numpy as np


def expScore(adata, geneSetScore):

    geneset = np.array([geneSetScore.get(var, 0.0) for var in adata.var_names])
    # Create a 1D array for the given Gene Set
    # with respect to gene names and their indexes
    # from annData object. If the gene name from adata.var_names
    # is absent in the geneset, put 0 as the value of the index.

    return adata.X.dot(geneset)
    # Return the dot product of gene names and our array. Since the indexes of
    # both arrays point to the same gene names,
    # we can simply return the dot product.



from setuptools import setup, find_packages

with open("README_PACKAGE.md", "r") as f:
    desc = f.read()

setup(name="gsnetact", 
      
      version="0.0.4",

      description="The GSNetAct Python Package.",
        
      long_description=desc,
    
      long_description_content_type="text/markdown",

      packages=find_packages(),
            
      )
      
